/* Copyright 2016-2021 Michele Santullo
 * This file is part of "duckhandy".
 *
 * "duckhandy" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "duckhandy" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "duckhandy".  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef id890F1A22A90047BE9A9DCC8F7F057F91
#define id890F1A22A90047BE9A9DCC8F7F057F91

#include "implem/int_conv.hpp"
#include <type_traits>
#include <string>
#include <string_view>
#include <stdexcept>

namespace dhandy {
	namespace implem {
		template <typename T, typename F, bool SafeRetval, typename Tr>
		struct IntConv;

		template <typename Tr>
		using IntConvString = std::basic_string<typename Tr::char_type>;

		template <typename Tr>
		using IntConvStringView = std::basic_string_view<typename Tr::char_type>;

		template <typename F, bool SafeRetval, typename Tr>
		struct IntConv<std::enable_if_t<std::is_integral_v<F>, IntConvString<Tr>>, F, SafeRetval, Tr> {
			static IntConvString<Tr> conv (const F& in) {
				auto retval = dhandy::int_to_ary<F, 10, Tr>(in);
				return IntConvString<Tr>{retval.begin(), retval.end() - 1};
			}
		};
		template <typename F, bool SafeRetval, typename Tr>
		struct IntConv<std::enable_if_t<std::is_integral_v<F>, IntConvStringView<Tr>>, F, SafeRetval, Tr> {
			constexpr static IntConvStringView<Tr> conv (const F& in) {
				if (std::is_constant_evaluated() or not SafeRetval)
					return dhandy::int_to_ary<F, 10, Tr>(in).to_string_view();
				else
					throw std::logic_error("Only callable in a constexpr context, call int_conv_temporary() instead if you know what you're doing");
			}
		};
		template <typename T, bool SafeRetval, typename Tr>
		struct IntConv<T, std::enable_if_t<std::is_integral_v<T>, IntConvString<Tr>>, SafeRetval, Tr> {
			constexpr static T conv (const IntConvString<Tr>& in) {
				return dhandy::ary_to_int<T, char, 10, Tr>(
					in.data(),
					in.data() + in.size()
				);
			}
		};
		template <typename T, bool SafeRetval, typename Tr>
		struct IntConv<T, std::enable_if_t<std::is_integral_v<T>, IntConvStringView<Tr>>, SafeRetval, Tr> {
			constexpr static T conv (const IntConvStringView<Tr>& in) {
				return dhandy::ary_to_int<T, char, 10, Tr>(
					in.data(),
					in.data() + in.size()
				);
			}
		};
		template <typename T, typename F, F In, bool SafeRetval, typename Tr>
		struct IntConv<T, std::integral_constant<F, In>, SafeRetval, Tr> {
			constexpr static T conv (const std::integral_constant<F, In>&) {
				return dhandy::buildtime_int_to_ary<F, In, 10, Tr>().to_string_view();
			}
		};
	} //namespace implem

	template <typename To, typename From>
	constexpr inline To int_conv (const From& from, bool all_caps=false) {
		if (all_caps)
			return implem::IntConv<To, From, true, ASCIITranslatorUpcase<char>>::conv(from);
		else
			return implem::IntConv<To, From, true, ASCIITranslator<char>>::conv(from);
	}

	template <typename To, typename From>
	constexpr inline To int_conv_temporary (const From& from) {
		return implem::IntConv<To, From, false, ASCIITranslator<char>>::conv(from);
	}
} //namespace dhandy

#endif
